#define _GNU_SOURCE

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include "cwebfiles.h"
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <errno.h>

extern char **environ;

// don't use before you've authenticated successfully!
struct cgi_serialized_session_status session;

// when we call this function, we usually leak memory - not an issue
static char *session_file_path(char *cookie) {
  char *file = malloc(strlen(SESSION_FOLDER)+COOKIE_LENGTH+1);
  memcpy(file, SESSION_FOLDER, strlen(SESSION_FOLDER));
  memcpy(file+strlen(SESSION_FOLDER), cookie, COOKIE_LENGTH);
  file[strlen(SESSION_FOLDER)+COOKIE_LENGTH] = '\0';
  return file;
}

int authenticate_by_cookie(char *cookie) {
  // check cookie string for conformity
  if (strlen(cookie) != COOKIE_LENGTH) return 1;
  if (!checkhex(cookie, COOKIE_LENGTH)) return 2;
  
  // read session data
  int fd = open(session_file_path(cookie), O_RDONLY);
  
  // verify that the session exists
  if (fd == -1) return 3;
  int res = read(fd, &session, sizeof(session));
  close(fd);
  
  // verify that the session file is valid
  if (res != sizeof(session)) return 4;
  session.user_name[32] = 0;
  
  // verify that the session is still active
  time_t cur_time = time(NULL);
  if (session.start_time + SESSION_LENGTH < cur_time) return 5;
  if (session.start_time > cur_time) return 6;
  
  // verify that the username and the userid still match
  struct passwd *user_data = getpwuid(session.uid);
  if (user_data == NULL) return 7;
  if (strcmp(user_data->pw_name, session.user_name) != 0) return 8;
  
  // everything seems fine! :)
  return 0;
}

void persist_session(char *cookie) {
  char *file = session_file_path(cookie);
  int fd = open(file, O_RDWR|O_EXCL|O_CREAT, S_IRUSR|S_IWUSR);
  if (fd == -1) senderr("can't create session file", MYFAULT);
  
  errno = 0;
  int res = write(fd, &session, sizeof(session));
  char errstr[128];
  if (res != sizeof(session)) {
    snprintf(errstr, 128, "can't write all session data (%i of %i): %s", (int)res, (int)sizeof(session), strerror(errno));
  }
  close(fd);
  if (res != sizeof(session)) {
    unlink(file);
    senderr(errstr, MYFAULT);
  }
}

void senderr(char *errmsg, bool myfault) {
  if (myfault) {
    puts("Status: 500 CGI error");
  } else {
    puts("Status: 400 CGI error");
  }
  puts("Content-Type: text/plain;charset=utf8");
  puts("X-Frame-Options: DENY");
  puts("");
  puts("An error occured while processing your request:");
  puts(errmsg);
  exit(0);
}

// may send an error response and exit
void grabrand(unsigned char *out, size_t len) {
  int randfd = open(RANDDEV, O_RDONLY);
  if (randfd == -1) senderr("can't open random device", MYFAULT);
  if (read(randfd, out, len) != len) senderr("reading random data failed", MYFAULT);
  close(randfd);
}

void login_and_setup() {
  char *cookies = getenv("HTTP_COOKIE");
  environ = NULL;
  setuid(0);
  
  // Note that this doesn't validate whether this is actually a cookie name or
  // part of a value - but that shouldn't be a problem.
  // Note: Redo if I feel like doing it.
  if (cookies == NULL) senderr("no cookies sent", NOTMYFAULT);
  char *session_cookie = strstr(cookies, COOKIE_NAME"=");
  if (session_cookie == NULL) senderr("session cookie not present", NOTMYFAULT);
  session_cookie += strlen(COOKIE_NAME"=");
  if (strlen(session_cookie) < COOKIE_LENGTH) senderr("invalid session cookie (error code 1)", NOTMYFAULT);
  session_cookie[COOKIE_LENGTH] = '\0';
  int res = authenticate_by_cookie(session_cookie);
  if (res != 0) {
    char errstr[128];
    snprintf(errstr, 128, "invalid session cookie (error code 2-%i)", res);
    senderr(errstr, NOTMYFAULT);
  }
}
